/***************************************************************************
*
* Copyright (c) 1989, 1990, 1991, 1992, 1993
* by Helsinki University of Technology, Digital Systems Laboratory.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  dynam.h
*
* Part of: PROD's programs "gphgen", probe, prpp, strong, and t
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

#define NODE  char
#define STACK LIST
#define QUEUE LIST
#define SET   LIST

typedef struct {
  NODE *(*copy)();
  void (*discard)(),(*dump)();
  BOOL (*better)();
} TOOLS;

typedef struct _LISTNODE {
  struct _LISTNODE *next;
  NODE *node;
} LISTNODE;

typedef struct {
  LISTNODE *first,*this,*last;
  TOOLS *tools;
} LIST;

typedef NODE *(*CFUNC)();

/* Initialisation of Dynamics */
extern void DynamInit();

/* General */
extern int Compare();
extern char *Copy();
extern void DumpIndent();

/* String Routines and Macros */
extern char *strcpy();
#define StrCopy(string) strcpy(Allocate((unsigned)(strlen(string) + 1)),string)
extern char *StrCp();
#define StrDiscard(string) Deallocate(string)
extern void StrDump();
extern BOOL StrBetter();
extern TOOLS strTools;

/* List Routines and Macros */
extern NODE *LstBackstep();
#define LstBegin(list) ((list)->this = (list)->first, LstThis(list))
#define LstBest(list) LstFirst(list)
extern LIST *LstCatenate();
extern LIST *LstCopy();
#define LstCorr(list1,list2,node2) LstNth(list1,LstWhich(list2,node2))
extern LIST *LstCreate();
extern NODE *LstDelete();
extern NODE *LstDelFirst();
extern NODE *LstDelKeyed();
extern NODE *LstDelLast();
extern NODE *LstDelNext();
extern NODE *LstDelOrdered();
extern void LstDestroy();
extern LIST *LstCp();
extern void LstDc();
extern void LstDump();
#define LstEmpty(list) (!(list)->first)
#define LstEnd(list) ((list)->this = (list)->last, LstThis(list))
extern NODE *LstFind();
#define LstFirst(list) ((list)->first ? (list)->first->node : NULL)
extern NODE *LstFndKeyed();
extern NODE *LstFndOrdered();
#define LstGet(list) ((char *) (list)->this)
extern NODE *LstInsFirst();
extern NODE *LstInsKeyed();
extern NODE *LstInsLast();
#define LstInsert(list,node) LstInsLast(list,node)
extern NODE *LstInsNext();
extern NODE *LstInsOrdered();
#define LstLast(list) ((list)->last ? (list)->last->node : NULL)
#define LstMember(list,node) (LstWhich(list,node) != 0)
extern NODE *LstNBegin();
#define LstNext(list) \
  ((list)->this->next ? (list)->this->next->node : NULL)
extern NODE *LstNth();
extern NODE *LstPrevious();
#define LstRelink(list,node) ((list)->this->node = (node))
extern NODE *LstRestore();
#define LstSet(list,pos) ((list)->this = (LISTNODE *) (pos))
extern unsigned LstSize();
#define LstStart(list) ((list)->this = NULL)
#define LstStep(list) \
  ((list)->this = ((list)->this ? (list)->this->next \
    : (list)->first), LstThis(list))
extern void LstStore();
#define LstThis(list) ((list)->this ? (list)->this->node : NULL)
extern unsigned LstWhich();
#define LstWorse(list) LstNext(list)
extern TOOLS lstTools;

/* Stack Macros */
#define StkCopy(stack,trueCopy)   LstCopy(stack,trueCopy)
#define StkCreate(tools)          LstCreate(tools)
#define StkDump(stack,depth)      LstDump(stack,depth)
#define StkDestroy(stack,discard) LstDestroy(stack,discard)
#define StkEmpty(stack)           LstEmpty(stack)
#define StkPop(stack)             LstDelFirst(stack,FALSE)
#define StkPush(stack,node)       LstInsFirst(stack,node)
#define StkSize(stack)            LstSize(stack)
#define StkTop(stack)             LstFirst(stack)
#define stkTools lstTools

/* Queue Macros */
#define QueCopy(queue,trueCopy)   LstCopy(queue,trueCopy)
#define QueCreate(tools)          LstCreate(tools)
#define QueDump(queue,depth)      LstDump(queue,depth)
#define QueDequeue(queue)         LstDelFirst(queue,FALSE)
#define QueDestroy(queue,discard) LstDestroy(queue,discard)
#define QueEmpty(queue)           LstEmpty(queue)
#define QueEnqueue(queue,node)    LstInsLast(queue,node)
#define QueFirst(queue)		  LstFirst(queue)
#define QueSize(queue)            LstSize(queue)
#define queTools lstTools

/* Set Macros and Routines */
extern SET *SetAdd();
#define SetBegin(set)             LstBegin(set)
#define SetBelongs(set,node)      (LstFind(set,node) != NULL)
#define SetCopy(set,trueCopy)     LstCopy(set,trueCopy)
#define SetCreate(tools)	  LstCreate(tools)
#define SetDelete(set,node,discard) LstDelete(set,node,discard)
#define SetDestroy(set,discard)   LstDestroy(set,discard)
extern SET *SetDifference();
#define SetDump(set,depth)        LstDump(set,depth)
#define SetEmpty(set)             LstEmpty(set)
extern SET *SetIntersection();
extern SET *SetSelect();
#define SetSize(set)              LstSize(set)
#define SetStart(set)             LstStart(set)
#define SetStep(set)              LstStep(set)
extern SET *SetSubtract();
extern SET *SetUnion();
#define SetWhich(set,node)        LstWhich(set,node)
#define setTools lstTools

/* Abbreviations */
#define LSTFOR(list,node,type) \
  for ((node) = (type *) LstBegin(list); \
       (node); \
       (node) = (type *) LstStep(list))
#define LSTNEST(list,node,type) \
  for (LstStart(list), LstStore(list), (node) = (type *) LstStep(list); \
       (node); \
       LstStore(list), (node) = (type *) LstStep(list))
#define LSTNUM(list,node,type,num) \
  for ((node) = (type *) LstBegin(list), (num) = 0; \
       (node); \
       (node) = (type *) LstStep(list), (num)++)
#define LSTROLL(list,node,type,num) \
  for (LstStart(list), LstStore(list), \
       (node) = (type *) LstStep(list), (num) = 0; \
       (node); \
       LstStore(list), (node) = (type *) LstStep(list), (num)++)
#define LSTUNNEST(list,node,type) \
  for ((void) LstRestore(list), (node) = (type *) LstThis(list); \
       (node); \
       (void) LstRestore(list), (node) = (type *) LstThis(list))
#define LSTUNROLL(list,node,type,num) \
  for ((num) = LstSize(list) - 1, (void) LstRestore(list), \
       (node) = (type *) LstThis(list); \
       (node); \
       (void) LstRestore(list), (node) = (type *) LstThis(list), (num)--)
