/***************************************************************************
*
* Copyright (c) 1989, 1990, 1991, 1992, 1993
* by Helsinki University of Technology, Digital Systems Laboratory.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet.h
*
* Part of: PROD's program "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     .
*
* Version history:
*     $Log$
***************************************************************************/

typedef struct PLC
{
    unsigned long count;
    LIST *pureIn;
    LIST *pureOut;
    LIST *neutral;
    LIST *mostIn;
    LIST *mostOut;
    LIST *orPredAnd;
    unsigned flag;
} PLC;

#define PlcAlloc() ((PLC *) Allocate(sizeof(PLC)))
extern PLC *PlcCons();
extern PLC *PlcCopy();
extern void PlcDiscard();
extern void PlcDump();
extern TOOLS plcTools;
extern void PlcClear();
extern void PlcWrite();
extern void PlcIniFlag();
extern void PlcAOClean();

typedef struct TRN
{
    unsigned flag;
    LIST *pureIn;
    LIST *pureOut;
    LIST *neutral;
    LIST *mostIn;
    LIST *mostOut;
    union
	{
	    LIST *higher;
	    unsigned long val;
	    struct
		{
		    LIST *plcPredAO;
		    LIST *orStrongPredAO;
		    LIST *orWeakPredAO;
		    unsigned long nOrSucc;
		} s;
	} u;
} TRN;

#define TrnAlloc() ((TRN *) Allocate(sizeof(TRN)))
extern TRN *TrnCons();
extern TRN *TrnCopy();
extern void TrnDiscard();
extern void TrnDump();
extern TOOLS trnTools;
extern void TrnRun();
extern void TrnEnblCheck();
extern void TrnAntiVisit();
extern void TrnCntEnbl();
extern void PtSRandom();
extern unsigned long PtTerm();
extern unsigned long PtRandom();
extern unsigned long TrnVisit();
extern void TrnStub();
extern void TrnBest();
extern void TrnPrecCheck();
extern void TrnReCheck();
extern void TrnAO();
extern void TrnAndOr();
extern void TrnCntCand();
extern void TrnCntStrong();
extern void TrnTryDel();
extern void TrnUndel();
extern void TrnDel();
extern void TrnAOClean();
extern void TrnDelCheck();

typedef struct PLARC
{
    TRN *trn;
    unsigned long weight;
} PLARC;

#define PlArcAlloc() ((PLARC *) Allocate(sizeof(PLARC)))
extern PLARC *PlArcCons();
extern PLARC *PlArcCopy();
extern void PlArcDump();
extern TOOLS plArcTools;

typedef struct PLDRC
{
    TRN *trn;
    unsigned long in;
    unsigned long out;
} PLDRC;

#define PlDrcAlloc() ((PLDRC *) Allocate(sizeof(PLDRC)))
extern PLDRC *PlDrcCons();
extern PLDRC *PlDrcCopy();
extern void PlDrcDump();
extern TOOLS plDrcTools;

typedef struct TRARC
{
    PLC *plc;
    unsigned long weight;
} TRARC;

#define TrArcAlloc() ((TRARC *) Allocate(sizeof(TRARC)))
extern TRARC *TrArcCons();
extern TRARC *TrArcCopy();
extern void TrArcDump();
extern TOOLS trArcTools;

typedef struct TRDRC
{
    PLC *plc;
    unsigned long in;
    unsigned long out;
} TRDRC;

#define TrDrcAlloc() ((TRDRC *) Allocate(sizeof(TRDRC)))
extern TRDRC *TrDrcCons();
extern TRDRC *TrDrcCopy();
extern void TrDrcDump();
extern TOOLS trDrcTools;

typedef struct PLAVL
{
    DATATYPE *key;
    PLC *plc;
    int balance;
    struct PLAVL *left;
    struct PLAVL *right;
} PLAVL;

#define PlAvlAlloc() ((PLAVL *) Allocate(sizeof(PLAVL)))
extern PLAVL *PlAvlCons();
extern PLAVL *PlAvlCopy();
extern void PlAvlDiscard();
extern void PlAvlDump();
extern TOOLS plAvlTools;

typedef struct TRAVL
{
    DATATYPE *key;
    TRN *trn;
    int balance;
    struct TRAVL *left;
    struct TRAVL *right;
} TRAVL;

#define TrAvlAlloc() ((TRAVL *) Allocate(sizeof(TRAVL)))
extern TRAVL *TrAvlCons();
extern TRAVL *TrAvlCopy();
extern void TrAvlDiscard();
extern void TrAvlDump();
extern TOOLS trAvlTools;
extern void TrAvlPrec();

typedef struct BINOR
{
    unsigned flag;
    TRN *trn;
} BINOR;

#define BinOrAlloc() ((BINOR *) Allocate(sizeof(BINOR)))
extern BINOR *BinOrCons();
extern BINOR *BinOrCopy();
#define BinOrDiscard(binOr) Deallocate((char *) (binOr))
extern void BinOrDump();
extern TOOLS binOrTools;

#define PlcGCount(plc) ((plc)->count)
#define PlcSCount(plc,_count) ((plc)->count = (_count))
#define PlcICount(plc,_count) ((plc)->count += (_count))
#define PlcDCount(plc,_count) ((plc)->count -= (_count))
#define PlcGPurIn(plc) ((plc)->pureIn)
#define PlcSPurIn(plc) ((plc)->pureIn = LstCreate(&plArcTools))
#define PlcDPurIn(plc) (LstDestroy((plc)->pureIn, TRUE), (plc)->pureIn = NULL)
#define PlcGPurOut(plc) ((plc)->pureOut)
#define PlcSPurOut(plc) ((plc)->pureOut = LstCreate(&plArcTools))
#define PlcDPurOut(plc) \
    (LstDestroy((plc)->pureOut, TRUE), (plc)->pureOut = NULL)
#define PlcGNeutral(plc) ((plc)->neutral)
#define PlcSNeutral(plc) ((plc)->neutral = LstCreate(&plArcTools))
#define PlcGMosIn(plc) ((plc)->mostIn)
#define PlcSMosIn(plc) ((plc)->mostIn = LstCreate(&plDrcTools))
#define PlcGMosOut(plc) ((plc)->mostOut)
#define PlcSMosOut(plc) ((plc)->mostOut = LstCreate(&plDrcTools))
#define PlcGOrPredAnd(plc) ((plc)->orPredAnd)
#define PlcSOrPredAnd(plc) ((plc)->orPredAnd = LstCreate(&trnTools))
#define PlcDOrPredAnd(plc) \
    (LstDestroy((plc)->orPredAnd, FALSE), (plc)->orPredAnd = NULL)
#define PlcSatisfied(plc) ((plc)->flag & 1)
#define PlcFeed(plc) ((plc)->flag |= 1)
#define PlcThirsty(plc) ((plc)->flag & 2)
#define PlcSalt(plc) ((plc)->flag |= 2)
#define PlcWater(plc) ((plc)->flag &= ~2)
#define PlcReset(plc) ((plc)->flag &= ~3)

#define TrnEnabled(trn) ((trn)->flag & 1)
#define TrnAccept(trn) ((trn)->flag |= 1)
#define TrnSatisfied(trn) ((trn)->flag & 2)
#define TrnFeed(trn) ((trn)->flag |= 2)
#define TrnThirsty(trn) ((trn)->flag & 4)
#define TrnSalt(trn) ((trn)->flag |= 4)
#define TrnWater(trn) ((trn)->flag &= ~4)
#define TrnAsleep(trn) ((trn)->flag & 8)
#define TrnPutToSleep(trn) ((trn)->flag |= 8)
#define TrnOld(trn) ((trn)->flag & 16)
#define TrnEnter(trn) ((trn)->flag |= 16)
#define TrnReset(trn) ((trn)->flag &= ~31)
#define TrnIsFact(trn) ((trn)->flag & 32)
#define TrnSIsFact(trn) ((trn)->flag |= 32)
#define TrnGPurIn(trn) ((trn)->pureIn)
#define TrnSPurIn(trn) ((trn)->pureIn = LstCreate(&trArcTools))
#define TrnDPurIn(trn) (LstDestroy((trn)->pureIn, TRUE), (trn)->pureIn = NULL)
#define TrnGPurOut(trn) ((trn)->pureOut)
#define TrnSPurOut(trn) ((trn)->pureOut = LstCreate(&trArcTools))
#define TrnDPurOut(trn) \
    (LstDestroy((trn)->pureOut, TRUE), (trn)->pureOut = NULL)
#define TrnGNeutral(trn) ((trn)->neutral)
#define TrnSNeutral(trn) ((trn)->neutral = LstCreate(&trArcTools))
#define TrnGMosIn(trn) ((trn)->mostIn)
#define TrnSMosIn(trn) ((trn)->mostIn = LstCreate(&trDrcTools))
#define TrnGMosOut(trn) ((trn)->mostOut)
#define TrnSMosOut(trn) ((trn)->mostOut = LstCreate(&trDrcTools))
#define TrnGHigher(trn) ((trn)->u.higher)
#define TrnSHigher(trn) ((trn)->u.higher = LstCreate(&trnTools))
#define TrnGVal(trn) ((trn)->u.val)
#define TrnSVal(trn,_val) ((trn)->u.val = (_val))
#define TrnGNOrSucc(trn) ((trn)->u.s.nOrSucc)
#define TrnCNOrSucc(trn) (((trn)->u.s.nOrSucc) = 0L)
#define TrnINOrSucc(trn) (((trn)->u.s.nOrSucc)++)
#define TrnDNOrSucc(trn) (((trn)->u.s.nOrSucc)--)
#define TrnGPlcPredAO(trn) ((trn)->u.s.plcPredAO)
#define TrnSPlcPredAO(trn) ((trn)->u.s.plcPredAO = LstCreate(&plcTools))
#define TrnDPlcPredAO(trn) \
    (LstDestroy((trn)->u.s.plcPredAO, FALSE), (trn)->u.s.plcPredAO = NULL)
#define TrnGOrStrongPredAO(trn) ((trn)->u.s.orStrongPredAO)
#define TrnSOrStrongPredAO(trn) \
    ((trn)->u.s.orStrongPredAO = LstCreate(&binOrTools))
#define TrnDOrStrongPredAO(trn) \
    (LstDestroy((trn)->u.s.orStrongPredAO, FALSE), \
     (trn)->u.s.orStrongPredAO = NULL)
#define TrnGOrWeakPredAO(trn) ((trn)->u.s.orWeakPredAO)
#define TrnSOrWeakPredAO(trn) \
    ((trn)->u.s.orWeakPredAO = LstCreate(&binOrTools))
#define TrnDOrWeakPredAO(trn) \
    (LstDestroy((trn)->u.s.orWeakPredAO, FALSE), \
     (trn)->u.s.orWeakPredAO = NULL)

#define PlArcGTrn(plArc) ((plArc)->trn)
#define PlArcGWeight(plArc) ((plArc)->weight)
#define PlArcIWeight(plArc,_weight) ((plArc)->weight += (_weight))

#define PlDrcGTrn(plDrc) ((plDrc)->trn)
#define PlDrcGIn(plDrc) ((plDrc)->in)
#define PlDrcGOut(plDrc) ((plDrc)->out)

#define TrArcGPlc(trArc) ((trArc)->plc)
#define TrArcGWeight(trArc) ((trArc)->weight)
#define TrArcIWeight(trArc,_weight) ((trArc)->weight += (_weight))

#define TrDrcGPlc(trDrc) ((trDrc)->plc)
#define TrDrcGIn(trDrc) ((trDrc)->in)
#define TrDrcGOut(trDrc) ((trDrc)->out)

#define PlAvlGKey(plAvl) ((plAvl)->key)
#define PlAvlGPlc(plAvl) ((plAvl)->plc)
#define PlAvlCPlc(plAvl) ((plAvl)->plc = NULL)
#define PlAvlGBalance(plAvl) ((plAvl)->balance)
#define PlAvlSBalance(plAvl,_balance) ((plAvl)->balance = (_balance))
#define PlAvlGLeft(plAvl) ((plAvl)->left)
#define PlAvlSLeft(plAvl,_left) ((plAvl)->left = (_left))
#define PlAvlGRight(plAvl) ((plAvl)->right)
#define PlAvlSRight(plAvl,_right) ((plAvl)->right = (_right))

#define TrAvlGKey(trAvl) ((trAvl)->key)
#define TrAvlGTrn(trAvl) ((trAvl)->trn)
#define TrAvlCTrn(trAvl) ((trAvl)->trn = NULL)
#define TrAvlGBalance(trAvl) ((trAvl)->balance)
#define TrAvlSBalance(trAvl,_balance) ((trAvl)->balance = (_balance))
#define TrAvlGLeft(trAvl) ((trAvl)->left)
#define TrAvlSLeft(trAvl,_left) ((trAvl)->left = (_left))
#define TrAvlGRight(trAvl) ((trAvl)->right)
#define TrAvlSRight(trAvl,_right) ((trAvl)->right = (_right))

#define BinOrGStrong(binOr) ((binOr)->flag & 1)
#define BinOrSStrong(binOr) ((binOr)->flag |= 1)
#define BinOrCStrong(binOr) ((binOr)->flag &= ~1)
#define BinOrGWeak(binOr) ((binOr)->flag & 2)
#define BinOrSWeak(binOr) ((binOr)->flag |= 2)
#define BinOrCWeak(binOr) ((binOr)->flag &= ~2)
#define BinOrHStrongNot(binOr) ((binOr)->flag & 4)
#define BinOrTStrongNot(binOr) ((binOr)->flag |= 4)
#define BinOrDStrongNot(binOr) ((binOr)->flag &= ~4)
#define BinOrHWeakNot(binOr) ((binOr)->flag & 8)
#define BinOrTWeakNot(binOr) ((binOr)->flag |= 8)
#define BinOrDWeakNot(binOr) ((binOr)->flag &= ~8)
#define BinOrGTrn(binOr) ((binOr)->trn)

extern unsigned long _nPlc;
#define NGPlc() _nPlc
#define NIPlc() (_nPlc++)
extern PTRBL *_tblPlc;
#define TblPlc() _tblPlc
extern unsigned long *_nTrn;
#define NGTrn(class) _nTrn[class]
#define NITrn(class) (_nTrn[class]++)
extern PTRBL **_tblTrn;
#define TblTrn(class) _tblTrn[class]
extern unsigned *_higher;
#define Higher(class) (_higher[class] & 1)
#define SetHigher(class) (_higher[class] |= 1)
#define SelfHigher(class) (_higher[class] & 2)
#define SetSelfHigher(class) (_higher[class] |= 3)
#define Lower(class) (_higher[class] & 4)
#define SetLower(class) (_higher[class] |= 4)
#define Dependent(class) _higher[class]
extern PLAVL *_avlPlc;
#define AvlPlc() _avlPlc
#define SetAvlPlc(plAvl) (_avlPlc = (plAvl))
extern TRAVL **_avlTrn;
#define AvlTrn(class) _avlTrn[class]
#define SetAvlTrn(class,trAvl) (_avlTrn[class] = (trAvl))
extern TRAVL *_lowTrAvl;
#define LowTrAvl() _lowTrAvl
#define SetLowTrAvl(trAvl) (_lowTrAvl = (trAvl))
extern unsigned _hiClass;
#define HiClass() _hiClass
#define SetHiClass(class) (_hiClass = (class))
extern BOOL _ptStub;
#define PtStub() _ptStub
#define SetPtStub() (_ptStub = TRUE)
#define ResetPtStub() (_ptStub = FALSE)
extern BOOL _plRnd;
#define PlRnd() _plRnd
#define SetPlRnd() (_plRnd = TRUE)
#define ResetPlRnd() (_plRnd = FALSE)
extern BOOL _ptLazy;
#define PtLazy() _ptLazy
#define SetPtLazy() (_ptLazy = TRUE)
#define ResetPtLazy() (_ptLazy = FALSE)
extern BOOL _trRnd;
#define TrRnd() _trRnd
#define SetTrRnd() (_trRnd = TRUE)
#define ResetTrRnd() (_trRnd = FALSE)
extern BOOL _ptDelAlg;
#define PtDelAlg() _ptDelAlg
#define SetPtDelAlg() (_ptDelAlg = TRUE)
#define ResetPtDelAlg() (_ptDelAlg = FALSE)
extern LIST *_binOrList;
#define BinOrList() _binOrList
#define BinOrSList() (_binOrList = LstCreate(&binOrTools))
#define BinOrDList() LstDestroy(_binOrList, TRUE)
extern LIST *_ptList;
#define PtList() _ptList
#define PtSList() (_ptList = LstCreate(&trnTools))
#define PtDList() LstDestroy(_ptList, FALSE)
extern unsigned long _ptSgCount;
#define PtSgCount() _ptSgCount
#define PtCSgCount() (_ptSgCount = 0L)
#define PtISgCount() (_ptSgCount++)
extern unsigned long _ptNow;
#define PtNow() _ptNow
#define PtCNow() (_ptNow = 0L)
#define PtINow() (_ptNow++)
extern unsigned long _ptBest;
#define PtBest() _ptBest
#define PtSetBest(b) (_ptBest = (b))
extern unsigned long _ptSzBest;
#define PtSzBest() _ptSzBest
#define PtSzSBest(sz) (_ptSzBest = (sz))
extern unsigned long _ptCounter;
#define PtSCounter(c) (_ptCounter = (c))
#define PtDCounter() (_ptCounter--)

extern void PtInit();
extern void AvlPlcDestroy();
extern void AvlTrnDestroy();
extern void PtEnd();
extern PLAVL *GetPlAvl();
extern TRAVL *GetTrAvl();
extern void PutInArc();
extern void PutOutArc();
extern void TrnReform();
extern void PlcFill();
extern void TrnFill();
extern unsigned PlcEncode();
extern void PlcDecode();
extern void PtSStub();
extern void PtFollowers();
extern void PtOutput();
